#!/usr/bin/python -u

#
# Ugobe Project Tool, using the ElementTree module (instead of the old PyXML)
#

# system modules
import sys
import string
import os.path
import shutil
import stat
import time
# this is used to forcably pull in the encoding we need for
# the freeze modules. ought to be a better way in the setup file
from encodings import ascii

try:
    #built into 2.4+
    import set
except ImportError:
    # otherwise try the pre-2.4 version
    from sets import Set as set

try:
    # built into 2.5
    from xml.etree import ElementTree as ET
except ImportError:
    # otherwise, need to try the external version
    from elementtree import ElementTree as ET
    # this is needed for IronPython
    #from elementtree import SimpleXMLTreeBuilder as TB

# local modules
from errorhandler import *
from xmlmacro import *

SOUND_BASE = 0x1000
MOTION_BASE = 0x2000
COMMAND_BASE = 0x3000
SCRIPT_BASE = 0x4000
PROPERTY_BASE = 0x5000

class UgobeProjectTool:
    def __init__(self):
        self.missing_commands = set()
        self.missing_sounds = set()
        self.missing_motions = set()
        self.missing_properties = set()
        self.enums = {}
        self.cmd_enums = None

    def str2bool(self,str):
        if str:
            if str.lower()=="true": return True
            if str.lower()=="yes":  return True
            else:
                try:
                    if int(rnd) == 0: return True
                except:
                    return False
        else:
            return False

    def get_nodeattr(self, node, attr):
        option = self.doc.find("/%s" % node)
        if isinstance(option, ET._Element):
            return option.get(attr)
        else:
            return None

    def get_option(self, name, default="", attr="value"):
        option = self.doc.find("/options/%s" % name)
        if isinstance(option, ET._Element):
            return option.get(attr)
        else:
            return default

    def get_bool_option(self, name):
        option = self.doc.find("/options/%s" % name)
        if isinstance(option, ET._Element):
            return True
        return False

    def get_directory(self, name):
        result = self.doc.find("*/directories/%s/" % name)
        if not isinstance(result, ET._Element):
            #print("directory %s not specified in XML file" % name)
            return None

        return result.get("value")

    def remove_directory(self, name):
        directory = self.get_directory(name)

        if os.path.isdir(directory):
            # this can cause issues if it is checked into svn
            shutil.rmtree(directory, True)

    def ensure_directory(self, name):
        directory = self.get_directory(name)
        temp_dir = ""

        if directory:
            for dir in directory.split("/"):
                temp_dir += "%s/" % dir
                if not os.path.exists(temp_dir):
                    os.mkdir(temp_dir)

            if not os.path.isdir(directory):
                error("%s is not a valid directory" % directory)

        return directory

    def get_path(self, name, path, extension = None):
        directory = self.get_directory(name)
        newPath = "%s/%s" % (directory, os.path.basename(path))

        if extension:
            (root, ext) = os.path.splitext(newPath)
            newPath = "%s.%s" % (root, extension)

        return newPath

    def get_numbered_path(self, path, resource, extension):
        directory = self.ensure_directory("build_%s" % resource)

        if not directory:
            directory = self.get_directory('build')
            if self.get_bool_option("folders"):
                if resource == "command":
                    directory = "%s/commands" % directory
                elif resource == "motion":
                    directory = "%s/motions" % directory
                elif resource == "sound":
                    directory = "%s/sounds" % directory
                elif resource == "script":
                    directory = "%s/scripts" % directory
                if not os.path.exists(directory):
                    os.mkdir(directory)

        (root, ext) = os.path.splitext(os.path.basename(path))
        eid = self.enums[resource][root]

        return "%s/%s.%s" % (directory, eid, extension)

    def check_exists(self, path):
        if not os.path.exists(path):
            error("file '%s' does not exist" % path)

    def remove_file(self, path):
        if os.path.exists(path):
            message("Removing %s" % path)
            os.remove(path)

    def copy_file(self, src, dst):
        self.check_exists(src)
        message("Copying %s to %s" % (src, dst))
        shutil.copyfile(src, dst)

    def file_depend(self, target, dependency):
        if not os.path.exists(target):
            return True
        if not os.path.exists(dependency):
            return False

        if os.stat(target)[stat.ST_MTIME] < os.stat(dependency)[stat.ST_MTIME]:
            return True

    def pre_process(self):
        # find all the sets
        sets = self.doc.findall("//set")
        # then set each one
        for s in sets:
            if s.get("name") == "COMMANDS":
                commandpath = s.get("value")
            elif s.get("name") == "MOTIONS":
                motionpath = s.get("value")
            elif s.get("name") == "SOUNDS":
                soundpath = s.get("value")

        strcmd = ""
        strmot = ""
        strsnd = ""

        op = file(opfilename,"w")
        op.write("\n")

        for strfilename in os.listdir(commandpath):
            if strfilename.endswith('csv'):
                op.write("    <command path=\"${COMMANDS}/%s\"/>\n" % strfilename)
        op.write("\n")

        for strfilename in os.listdir(motionpath):
            if strfilename.endswith('csv'):
                op.write("    <motion path=\"${MOTIONS}/%s\"/>\n" % strfilename)
        op.write("\n")

        for strfilename in os.listdir(soundpath):
            if strfilename.endswith('wav'):
                op.write("    <sound path=\"${SOUNDS}/%s\"/>\n" % strfilename)

        op.close()

    def killdups(self, some_list):
        nodups = {}
        l = []
        for x in some_list:
            nodups[x] = 1
        l = nodups.keys()
        l.sort()
        return l

    def is_num(self, strfilename):
        try:
            n = int(strfilename)
            return True
        except:
            return False

    def pre_process_cmdstyle(self):
        print "initializing..."

        import csv2ucf
        import csv2umf
        strmot = ""
        strsnd = ""
        motcnt = 0
        sndcnt = 0
        m2u = csv2umf.Motion()
        c2u = csv2ucf.Commands()
        c2u.set_include(".:" + self.get_option("include"))

        sets = self.doc.findall("//set")
        for s in sets:
            if s.get("name") == "MOTIONS":
                motionpath = s.get("value")
            elif s.get("name") == "SOUNDS":
                soundpath = s.get("value")

        print "loading commands..."
        commands = self.doc.findall("//resources.command")
        for command in commands:
            path = command.get("path")
            if os.path.exists(path):
                c2u.load_csv(path)
            else:
                print "ERROR: COMMAND %s NOT FOUND" % path

        motionlist = []

        print "extracting motions..."
        for item in c2u.entries:
            motionlist.append(item['motion'])

        print "removing duplicate motions"
        noduplist = self.killdups(motionlist)

        print "loading motions "
        for motionname in noduplist:
            strfilename = "%s.csv" % motionname.strip()
            strmot += "    <motion path=\"${MOTIONS}/%s\"/>\n" % strfilename
            path = "%s/%s" % (motionpath, strfilename)
            if os.path.exists(path):
                m2u.read_from_csv(path)
            else:
                print "ERROR: MOTION %s NOT FOUND" % path
            motcnt += 1

        print "removing duplicate sounds"
        noduplist = self.killdups(m2u.data['SD'])

        print "extracting sounds"
        for strfilename in noduplist:
            strfilename = strfilename.strip()
            path = "%s/%s.wav" % (soundpath, strfilename)
            if os.path.exists(path):
                strsnd += "    <sound path=\"${SOUNDS}/%s.wav\"/>\n" % strfilename
            else:
                print "ERROR: SOUND %s NOT FOUND" % path
            sndcnt += 1

        #filename known from start
        op = file(opfilename,"w")
        op.write(strmot)
        op.write("\n")
        op.write(strsnd)
        op.close()
        print "wrote %d motions and %d sounds to %s" % (motcnt, sndcnt, opfilename)

    def execute(self, cmd, inFile, outFile):
        command = self.doc.find("//%s" % cmd)
        #print command
        #command = xpath.Evaluate("/ugobe_project/options/tools/%s[@value]" % cmd, self.doc)

        if not ET.iselement(command):
            error("build command '%s' not found in project file" % cmd)

        command = command.get("value")
        command = os.path.normpath(command)
        #command = command[0].getAttribute("value")
        command = command.replace('%i', inFile)
        command = command.replace('%o', outFile)

        includes = ""
        #print "includes:", self.get_option('include')
        for i in string.split(self.get_option('include'), ':'):
            includes = "%s -i%s" % (includes, os.path.normpath(i))
        #print includes
        command = command.replace('%I', includes)

        message("@ %s" % command)

        return os.system(command) == 0

    def build_script(self, resource):
        path = resource.get("path")
        opts = self.get_option("script_build")

        # If this is one of the 'special' names, leave it. Otherwise,
        # do the renumbering business
        #print "path is", path
        if (path == "main.p") or (path == "sensors.p") or (path == "init.p") or (path == "startup.p"):
            #print "building to root"
            dPath = self.get_path("build", path, "amx")
        else:
            #print "building to scripts folder"
            dPath = self.get_numbered_path(path, 'script', 'amx')

        self.check_exists(path)

        if self.options.noscriptrebuild:
            if not (self.file_depend(dPath, path)):
                #message("not rebuilding script %s: no changes detected" % path)
                return dPath

        #message("Building script '%s'" % path)
        if not self.execute("pawn", path, dPath):
            error("building script '%s'" % path)

        if self.get_bool_option("block"):
            self.options.block = True

        if self.options.block:
            message("block compiling '%s'" % dPath)
            if not self.execute("block", dPath, dPath):
                error("block compiling '%s'" % dPath)

        return dPath

    def build_motion(self, resource):
        path = resource.get("path")
        dPath = self.get_numbered_path(path, 'motion', 'umf')

        if not (self.file_depend(dPath, path) or self.file_depend(dPath, "sounds.xml")):
            return dPath

        (root, ext) = os.path.splitext(path)

        if ext == ".umf":
            self.copy_file(path, dPath)

        elif ext == ".csv":
            target = self.get_option("target", "pleo")

            import csv2umf
            motion = csv2umf.Motion(target)

            self.check_exists(path)

            if 'sound' in self.enums:
                sounds = self.enums['sound']
            else:
                sounds = None

            if 'command' in self.enums:
                commands = self.enums['command']
            else:
                commands = None

            # what UMF version do we want?
            motion_version = self.get_option("motion", None, "version")
            if motion_version is None:
                motion_version = self.get_option("umf")
            if motion_version == "":
                motion_version = 3
            else:
                motion_version = int(motion_version)
                
            #do NOT feed this an fp. it needs the filename
            motion.read_from_csv(path)
            
            if motion_version == 2:
                message("Converting %s to %s, UMF2 format" % (path, dPath))
                motion.write_to_umf2(file(dPath, "wb"), sounds, commands)
            elif motion_version == 3:
                message("Converting %s to %s, UMF3 format" % (path, dPath))
                motion.write_to_umf3(file(dPath, "wb"), sounds, commands)
            else:
                message("Unknown UMF version asked for")

            self.missing_sounds = (self.missing_sounds.union(motion.missing_resources))

        else:
            error("Cannot convert motion file %s of this type" % path)

        return dPath


    def build_command(self, resource):
        path = resource.get("path")
        dPath = self.get_numbered_path(path, 'command', 'ucf')

        if not (self.file_depend(dPath, path) or self.file_depend(dPath, "motions.xml")):
            return dPath

        (root, ext) = os.path.splitext(path)

        if ext == ".ucf":
            self.copy_file(path, dPath)

        elif ext == ".csv":
            message("Converting %s to %s" % (path, dPath))

            target = self.get_option("target", "pleo")

            import csv2ucf
            csv2ucf = csv2ucf.Commands(target)
            csv2ucf.set_include(".:" + self.get_option("include"))

            if self.cmd_enums:
                csv2ucf.enums = self.cmd_enums
                csv2ucf.enums_loaded = True

            try:
                self.check_exists(path)
                csv2ucf.load_csv(path)
                csv2ucf.save_ucf(dPath)

                if not self.cmd_enums:
                    self.cmd_enums = csv2ucf.enums

                self.missing_motions = (self.missing_motions.union(csv2ucf.missing_resources))
                self.missing_properties = (self.missing_properties.union(csv2ucf.missing_properties))

            except MsgException, strerror:
                self.missing_commands.add(path)
                print(strerror)

        else:
            error("Cannot convert command file %s of this type" % path)

        return dPath

    def build_sound(self, resource):
        path = resource.get("path")

        dPath = self.get_numbered_path(path, 'sound', 'usf')

        if not self.file_depend(dPath, path):
            return dPath

        # Should we ADPCM encode it?
        if self.get_option("sound", None, "adpcm") == "true":
            adpcm = True
        else:
            if self.get_option("adpcm") == "true":  # allow global option to set all true
                adpcm = True
            else:
                adpcm = False

        adpcm_local = resource.get("adpcm") # allow local to override global
        if adpcm_local == "true":
            adpcm = True
        elif adpcm_local == "false":
            adpcm = False

        # Can we pitch shift this sound?
        # currently, we accept yes|no, true|false. should we also accept integer values?
        pitch_local = resource.get("pitch", "true")
        if pitch_local == "yes" or pitch_local == "true":
            allow_pitch = True
        else:
            allow_pitch = False

        # Get the frequency we should use. Default to 11025
        frequency = self.get_option("sound", 11025, "frequency")
        
        frequency_local = resource.get("frequency") # allow local to override global
        if frequency_local is not None:
            frequency = frequency_local

        if frequency is None:
            freq = 11025
        else:
            freq = int(frequency)

        try:
            self.check_exists(path)

            (root, ext) = os.path.splitext(path)

            if ext == ".usf":
                self.copy_file(path, dPath)

            elif ext == ".wav":
                import wav2usf
                
                width = 1
                channels = 1

                message("Converting %s to %s adpcm:%d pitch:%d freq:%d" % (path, dPath, adpcm, allow_pitch, freq))

                tab_up()
                (freq, width, channels, samples) = wav2usf.convert(path, freq, width, channels, adpcm)
                wav2usf.write_usf(samples, width, freq, channels, 1, dPath, path, adpcm, allow_pitch)
                tab_down()

            else:
                error("Cannot convert sound file %s of this type" % path)
        except MsgException, strerror:
                self.missing_sounds.add(path)
                print(strerror)

        return dPath

    def build_file(self, resource):
        path = resource.get("path")
        dPath = self.get_path("build", path)

        if self.file_depend(dPath, path):
            self.copy_file(path, dPath)

        return dPath

    def get_method(self, name):
        if hasattr(self, name):
            method = getattr(self, name)

            # Python 3.0 no longer has the callable function, so
            # we need to do this:
            #if hasattr(method, '__call__'):
            if callable(method):
                return method

        # warning("method '%s' not defined" % name)
        return None

    def build_enumeration(self, name, resource, short = None, minimum = 1, nonone = False):
        enum = None
        maximum = None
        randomids = False

        import enumupdate

        message("Updating enumeration XML '%s.xml'" % name)
        values = self.doc.findall("//resources/%s" % resource)

        #print "%s min: %d max: %d" % (resource, minimum, maximum)
        values = map(lambda x: os.path.splitext(os.path.basename(x.get("path")))[0], values)

        if len(values) > 0:
            enum = enumupdate.enumupdate("%s.xml" % name, resource, values,
                                         short = short, randomids = randomids,
                                         minimum = minimum, namesoff = True)

            filename = name
            directory = self.get_directory('include')
            if directory:
                filename = "%s/%s" % (directory, filename)

            inFilename = "%s.xml" % filename
            outFilename = "%s.inc" % filename

            if os.path.exists(outFilename):
                os.remove(outFilename)
            #if self.file_depend(outFilename, inFilename):
            import enumgen

            message("Creating enumeration '%s'" % outFilename)
            enumgen.enumgen(enum, file(outFilename, "wt"), nonone)

            self.enums[resource] = enum
        else:
            message("no data for %s" % name)

    def have_resource(self, name):
        resource = self.doc.find("//%s" % name)
        return resource is not None

    def cmd_build(self):
        message("\n*** Prepocessing ***")
        tab_up()

        # Create directories
        self.ensure_directory("include")
        self.ensure_directory("build")

        # Create enumerations
        self.build_enumeration('sounds', 'sound', 'snd', minimum = 0x1000)
        self.build_enumeration('motions', 'motion', 'mot', minimum = 0x2000)
        self.build_enumeration('commands', 'command', 'cmd', minimum = 0x3000)
        self.build_enumeration('scripts', 'script', 'scr', minimum = 0x4000)
        self.build_enumeration('user_properties', 'user_property', 'property', minimum = 0x5000, nonone = True)

        filelist = []
        resourcemap = []

        self.sounds = {}
        self.motions = {}
        self.commands = {}
        self.scripts = {}
        self.properties = {}

        message("\n*** Processing ***")

        resources = self.doc.find("//resources/")
        for resource in list(resources):

            # get the base name for the lookup file
            path = resource.get("path")
            (name, ext) = os.path.splitext(os.path.basename(path))
            #print name, ext

            method = self.get_method("build_%s" % resource.tag)
            if method:
                output_file_name = method(resource)
                if output_file_name:
                    (id, ext) = os.path.splitext(os.path.basename(output_file_name))
                    if resource.tag == "sound":
                        self.sounds[int(id)] = (name, output_file_name)
                    if resource.tag == "motion":
                        self.motions[int(id)] = (name, output_file_name)
                    if resource.tag == "command":
                        self.commands[int(id)] = (name, output_file_name)
                    if resource.tag == "script":
                        # we want to use the ID, even for the known names.
                        scripts = self.enums['script']
                        #print scripts
                        self.scripts[scripts[name]] = (name, output_file_name)
                else:
                    warning("resource %s did not produce an output file (was not compiled)" % resource.tag)
            else:
                # if there is a user_property enum, spit those out too
                if resource.tag == "user_property":
                    props = self.enums['user_property']
                    if props:
                        value = resource.get("value", "0")
                        self.properties[props[name]] = (name, int(value))

        self.write_res_file()

        tab_down()

    def write_res_file(self):

        root = self.doc.getroot()
        name = root.attrib['name']
        directory = self.get_directory('build')
        self.path = "%s/%s.urf" % (directory, name)
        
        message("\n*** Writing %s ***" % self.path)

        # create arrays of tuples for the write
        sounds = [ (k,self.sounds[k]) for k in sorted(self.sounds.keys())]
        motions = [ (k,self.motions[k]) for k in sorted(self.motions.keys())]
        commands = [ (k,self.commands[k]) for k in sorted(self.commands.keys())]
        scripts = [ (k,self.scripts[k]) for k in sorted(self.scripts.keys())]
        properties = [ (k,self.properties[k]) for k in sorted(self.properties.keys())]

        # structure to help us make more general code below
        resource_types = [
            # Base ID,    Signature, Alignment, Array of resource, TOC, start index in URF
#            [  SOUND_BASE,   'UGSF', 512,     sounds, [], 0],
            [  SOUND_BASE,   'UGSF',   4,     sounds, [], 0],
            [ MOTION_BASE,   'UGMF',   4,    motions, [], 0],
            [COMMAND_BASE,   'UGCF',   4,   commands, [], 0],
#            [ SCRIPT_BASE,   ' AMX', 512,    scripts, [], 0],
            [ SCRIPT_BASE,   ' AMX',   4,    scripts, [], 0],
            [PROPERTY_BASE,  'PROP',   4, properties, [], 0]
        ]

        # here we try to get the subversion ID, using the svnversion program
        import subprocess

        try:
            # first, let us see if there is a version in the UPF
            version_str = self.get_option("version")
            if version_str == "":
                # if not, try to get the Subversion revision
                p = subprocess.Popen(["svnversion"], stdout=subprocess.PIPE)
                version_str = p.communicate()[0]
            # make sure we have a valid integer i.e. no "numM" or "num:num"
            version = int(version_str.strip(string.letters+string.whitespace).rsplit(":")[-1])
        except:
            print "Bad version element in UPF or cannot get Subversion revision. Using 0"
            version = 0
        print "Version is %d" % (version)

        # build time
        build_time = int(time.time())
        print "Build Time is %d (%s)" % (build_time, time.ctime(build_time))

        # Now lets write out the binay res file
        import struct

        fout = file(self.path, "wb")

        # write the (16 byte) header
        fout.write(struct.pack('4s', 'UGRF'))         # ID
        fout.write(struct.pack('<I', 1))              # file format version
        fout.write(struct.pack('<I', build_time))     # time
        fout.write(struct.pack('<I', version))        # version from UPF or Subversion revision

        # put the place holders for the TOC offsets
        for type in resource_types:
            sig = type[1]
            start = type[5]
            fout.write(struct.pack('4s', sig))
            fout.write(struct.pack('<I', start))

        # and the placeholder for the SIZE record
        fout.write(struct.pack('4s', 'SIZE'))    # SIZE record
        fout.write(struct.pack('<I', 0))        # Total size of this resource file, minus CRC_SIG+CRC

        # now go through each resource type and dump them out
        for resource_type in resource_types:
            next = resource_type[0]
            resources = resource_type[3]
            toc = resource_type[4]

            # here we want to align on a 512 byte boundary
            ppos = fout.tell()
            npos = ((ppos + 511) & ~511) - ppos
            fout.write(npos*'0')

            # now go through each resource, saving to the drive, and updating the TOC
            if next == PROPERTY_BASE:
                for (id, (name, value)) in resources:
                    while next != id:
                        print "adding pad at %s" % (hex(next))
                        toc.append((0xFFFFFFFF, 0, ""))
                        next = next + 1
                    next = next + 1

                    # for properties, the data is the default value. perhaps this can be set in the upf
                    data = value
                    toc.append((fout.tell(),len(name),name))
                    print "writing %s (%s) at %s" % (name, hex(int(id)), hex(fout.tell()))
                    fout.write( struct.pack('<I', data) )
            else:
                for (id, (name, path)) in resources:
                    while next != id:
                        print "adding pad at %s" % (hex(next))
                        toc.append((0xFFFFFFFF, 0, ""))
                        next = next + 1
                    next = next + 1

                    fin = file(path, "rb")
                    if fin:
                        # here we want to align on a 512 byte boundary
                        ppos = fout.tell()
                        npos = ((ppos + resource_type[2]-1) & ~(resource_type[2]-1)) - ppos
                        fout.write(npos*'0')

                        data = fin.read()
                        toc.append((fout.tell(),len(data),name))
                        print "writing %s (%s) at %s" % (path, name, hex(fout.tell()))
                        fout.write(data)
                        fin.close()

        # here we want to align on a 512 byte boundary
        ppos = fout.tell()
        npos = ((ppos + 511) & ~511) - ppos
        fout.write(npos*'0')

        # now write out the TOCs
        for type in resource_types:
            sig = type[1]
            toc = type[4]
            start = type[5] = fout.tell()
            print "writing %s toc at %s" % (sig, hex(start))
            fout.write(struct.pack('4s', sig))
            fout.write(struct.pack('<I', len(toc)*40))
            for entry in toc:
                fout.write(struct.pack('<II32s', entry[0], entry[1], entry[2]))

        # pad the end with a few too
        ppos = fout.tell()
        npos = ((ppos + 7) & ~7) - ppos
        fout.write(npos*'0')

        # the length of the file does not include the ADLER sig or value
        urf_length = fout.tell()

        # now fix up the master TOC
        fout.seek(16)

        for type in resource_types:
            sig = type[1]
            start = type[5]
            fout.write(struct.pack('4s', sig))
            fout.write(struct.pack('<I', start))

        # write the size tag again with the proper value
        fout.write(struct.pack('4s', 'SIZE'))
        fout.write(struct.pack('<I', urf_length))

        fout.close()

        # check total length
        size_limit = 54 * 128 * 528
        if urf_length + 8 > size_limit:
            warning("URF file too large: %d of %d. %d over\n" % (urf_length+8, size_limit, (urf_length+8-size_limit)))
        else:
            message("URF file fits: %d of %d. %d free\n" % (urf_length+8, size_limit, size_limit-urf_length+8))

        # now calculate the adler32
        self.add_checksum()

    def add_checksum(self):
        import zlib
        import struct

        res_fp = file(self.path, "a+b")
        res_fp.seek(0)
        
        if res_fp:
            # look for an existing checksum
            crc = 0
            while 1:
                buf = res_fp.read(512)
                if len(buf) <= 0:
                    break
                crc = zlib.adler32(buf, crc)
            print "Adler32 crc is %X" % (crc & 0xFFFFFFFF)
            #res_fp.seek(0, os.SEEK_END)
            res_fp.write(struct.pack('4s', 'ADLR'))
            res_fp.write(struct.pack('<I', crc & 0xFFFFFFFF))
            # should we adjust the SIZE tag too?
            res_fp.close()
        # TODO: add error handler here

    def cmd_clean(self):
        message("\n*** Cleaning ***")
        tab_up()

        includeDirTmp = self.get_directory('include')
        if includeDirTmp:
            includeDir = "%s/" % includeDirTmp

        for i in ['sounds', 'motions', 'commands', 'scripts', 'user_properties']:
            self.remove_file("%s%s.inc" % (includeDir, i))
            self.remove_file("%s.xml" % (i))

        if includeDirTmp:
            if os.path.exists(includeDirTmp):
                shutil.rmtree(includeDirTmp)

        buildDir = self.get_directory('build')
        if os.path.exists(buildDir):
            if self.options.exclusive:
                message("Exclusive Clean: Removing commands only from '%s'" % buildDir)
                for f in os.listdir(buildDir):
                    if f.endswith('.ucf') or f.endswith('.amx'):
                        path = os.path.join(buildDir, f)
                        os.remove(path)

            else:
                message("Complete Clean: Removing build directory '%s'" % buildDir)
                shutil.rmtree(buildDir, True)

        tab_down()

    def cmd_rebuild(self):
        self.cmd_clean()
        self.cmd_build()

    def cmd_gui(self):
        error("GUI not yet implemented")

    def run(self, doc, options, command):
        self.doc = doc
        self.options = options

        if self.options.preprocess:
            self.pre_process_cmdstyle()
        else:
            method = self.get_method("cmd_%s" % command)
            if method:
                method()
            else:
                error("invalid command '%s'" % command)

            if (len(self.missing_sounds) or len(self.missing_motions) or len(self.missing_properties)):
                message("\nSummary of missing resources:")

                tab_up()

                for command in self.missing_commands:
                    warning("missing command '%s'" % command)

                for sound in self.missing_sounds:
                    warning("missing sound '%s'" % sound)

                for motion in self.missing_motions:
                    warning("missing motion '%s'" % motion)

                for prop in self.missing_properties:
                    warning("missing property '%s'" % prop)

                tab_down()


def help(stream, parser):
    parser.print_help(stream)

    stream.write("\ncommands:\n")
    stream.write("  build       Build the project.\n")
    stream.write("  clean       Remove generated project files.\n")
    stream.write("  rebuild     Build then clean.\n")
    stream.write("  gui         Load graphical user interface. (default)\n")


def main():
    # try to import and use psyco
    try:
        import psyco
        #psyco.log()
        #psyco.profile()
        psyco.full()
    except ImportError:
        pass

    from optparse import OptionParser, IndentedHelpFormatter
    
    if sys.platform == "darwin" or sys.platform == "mac":
        os.chdir(os.environ["PWD"])

    # Parse args
    parser = OptionParser(usage = "%prog [options] <project.upf> [command]", add_help_option = False)
    parser.add_option("-x", "--exclusive", dest="exclusive", help="exclusive rebuild", action="store_true")
    parser.add_option("-p", "--preprocess", dest="preprocess", help="preprocess cmds", action="store_true")
    parser.add_option("-b", "--block", dest="block", help="blockload", action="store_true")
    parser.add_option("-c", "--color", dest="color", help="turn on color output", action="store_true")
    parser.add_option("-h", "--help", dest="help", help="show this help message and exit", action="store_true")
    parser.add_option("-s", "--noscriptrebuild", dest="noscriptrebuild", help="do not rebuild scripts that haven't changed", action="store_true")
    (options, args) = parser.parse_args()

    # Process args
    if options.help:
        help(sys.stdout, parser)
        return 0

    if len(args) < 1 or 2 < len(args):
        help(sys.stderr, parser)
        return 1

    if options.color:
        errorhandler.color_mode = True

    filename = args[0]
    if len(args) == 2:
        if options.preprocess:
            opfilename = args[1]
            command = "none"
        else:
            command = args[1]
    else:
        help(sys.stderr, parser)
        return 1

    # Read project file
    if not os.path.exists(filename):
        if command != "gui":
            sys.stderr.write("project file '%s' does not exist\n" % filename)
            return 1

        doc = None

    else:
        # this is needed for IronPython
        #parser = TB.TreeBuilder()
        #doc = ET.parse(filename, parser)
        doc = ET.parse(filename)

        mp = MacroProcessor()
        mp.process(doc)

    app = UgobeProjectTool()

    try:
        t = time.time()
        app.run(doc, options, command)
        print "build time: %f sec" % (time.time() - t)

    except MsgException, x:
        sys.stderr.write(x.msg)
        return 1


if __name__ == "__main__":
    main()
